"""
CLI entry point, when used as a module: ``python -m kopf``.

Useful for debugging in the IDEs (use the start-mode "Module", module "kopf").
"""
from kopf import cli

if __name__ == '__main__':
    cli.main()
